' Jenco Instruments Serial Communications Protocol
' File           : DEVICES.BAS
' Date           : 1992,1997
' Quick Basic
' Implementation : RLL

DEFINT A-Z

'------------------
' Global Variables
'------------------
DIM SHARED address, Timeout, BaseAddress(1 TO 5), Bufferl&, Bufferb(0 TO 3)
DIM SHARED Bufferw&(0 TO 1), ComResult, DevMsg
DIM SHARED length, exponent, Signed, IdenNumber
DIM SHARED THR, RBR, LSR, LCR, MCR, IER, DLL, DLM, TRC

'-----------
' Constants
'-----------
CONST com1 = 1
CONST com2 = 2
CONST com3 = 3
CONST com4 = 4
CONST usercom = 5
CONST RTR = &HD0
CONST Nack = &H3E
CONST Errorfollows = &H3C
CONST Transmit = 2
CONST Receive = 1
CONST ReadD = 1  'old 0
CONST WriteD = 2  ' old 1
CONST InvokeCmd = 4
CONST ComnoErr = 1
CONST BaudErr = 2
CONST TimeErr = 3
CONST DevBusy = 4
CONST Badpara = 5
CONST BadCom = 6
CONST BadData = 7
CONST ResetErr = 8
CONST UserErr = 9
CONST NoCom = 10
CONST AddrErr = 11
CONST Ln10 = 2.302585093#
CONST Ln256 = 5.545177444#

'----------------------------------------
' Subroutine Declarations called by USER
'----------------------------------------
DECLARE SUB GetDevType (DevType&, Code%)
'Variable passing: DevType% by reference, Code% by reference
'-----------------
DECLARE SUB Initcom (Comx%, BR%, Code%)
'Variable passing:   Comx% by value, BR% by value, Code% by reference
'-----------------
DECLARE SUB Invoke (Cmd%, Code%)
'Variable passing:  Cmd% by value, Code% by reference
'-----------------
DECLARE SUB SetUserCom (addr%, Code%)
'Variable passing:      addr% by value, Code% by reference
'-----------------
DECLARE SUB SwitchtoCom (Comx%, Code%)
'Variable passing:       Comx% by value, Code% by reference
'-----------------
DECLARE SUB SwitchToAddress (address%, Code%)
'Variable passing:           address% by value, Code% by reference
'-----------------
DECLARE SUB ResetDevice (Code%)
'Variable passing:       Code% by reference
'-----------------
DECLARE SUB ReadData (Parameter&, datax#, Code%)
'Variable passing:    Parameter& by value, Datax# by reference, Code% by reference
'-----------------
'ReadFlags is not available in Quick Basic 4.5 due to the limitations of the
'compiler
'-----------------
DECLARE SUB ReadFlags (Parameter&, datax&, Code%)
'Variable passing:     Parameter& by value, Datax& by reference, Code% by reference
'-----------------
DECLARE SUB ReadFlag (Parameter&, bit%, datax%, Code%)
'Variable passing:    Parameter& by value, Datax% by reference, Code% by reference
'-----------------
DECLARE SUB WriteData (Parameter&, datax#, Code%)
'Variable passing:     Parameter& by value, Datax# by reference, Code% by reference
'-----------------

'----------------------------------------------------------------
' Subroutines not called by user but needed by above subroutines
'----------------------------------------------------------------
DECLARE SUB Initials (command%, Parameter&, Check%, Code%)
DECLARE SUB getaddress (zap%, adx%)
DECLARE SUB Talk (Speak%, Hear%)
DECLARE SUB ReadCom (datax%)
DECLARE SUB RG (command%, Parameter&, Check%, datax#, Code%)
DECLARE SUB WriteCom (datax%)
DECLARE SUB WG (Code%)

'-------------------
' Subroutine Errors
'-------------------
' Error code definitions
' -1  Com no. Error  [Valid 1 to 5]
' -2  Baud rate error [Valid 300,600,1200,2400,4800 & 9600]
' -3  TimeOut no response  [ No Response for more than 2 sec ]
' -4  Device Busy     [Parameter passed cannot be serviced at the moment]
' -5  Unknown parameter [Parameter not supported]
' -6  Invalid Command   [Command not supported ]
' -7  Invalid Data [ Data too big etc.]
' -9  RS485 Base Address Error [Base Address not supported by RS485 card]
'-10  No RS232/RS485 installed in selected com port
'-11  RS485 Address Error [Address < 0 or Address >255]
'
' For additional errors see man6171.txt.

'-------------------------
' Start of User's program
'-------------------------

'-----------------------
' End of user's program
'-----------------------

SUB getaddress (zap%, adx%)
    STATIC baseaddr%
    IF zap% = 0 THEN baseaddr% = adx% ELSE adx% = baseaddr%
END SUB

'----------------------
' GETDEVTYPE procedure
'----------------------
SUB GetDevType (DevType&, Code%)
    TryCmd% = 0
    DO
       OUT LCR%, &H2B
       Talk (address), DevTypex
       DevType& = DevTypex * 256
       OUT LCR%, &H3B
       IF ComResult <> 0 THEN
	  Code% = ComResult%
	  EXIT SUB
       END IF
       Talk (TryCmd%), Reply%
       IF Reply% <> Nack THEN
	  DevType& = DevType& OR Reply%
	  EXIT SUB
       END IF
       TryCmd% = TryCmd% + 1
    LOOP UNTIL TryCmd% > 5
    Code% = BadCom
END SUB

'-------------------
' INITCOM procedure
'-------------------
SUB Initcom (Comx, BR, Code%) 'Comx by value, BR by value
    IF (BR <> 300) AND (BR <> 600) AND (BR <> 1200) AND (BR <> 2400) AND (BR <> 4800) AND (BR <> 9600) THEN
       ComResult = BaudErr
       EXIT SUB
    END IF
    SwitchtoCom (Comx), Code%
    IF Code% <> 0 THEN EXIT SUB
    OUT LCR, &HBB
    SELECT CASE BR
       CASE 300
	  Datalsb = &H80
	  Datamsb = 1
       CASE 600
	  Datalsb = &HC0
	  Datamsb = 0
       CASE 1200
	  Datalsb = &H60
	  Datamsb = 0
       CASE 2400
	  Datalsb = &H30
	  Datamsb = 0
       CASE 4800
	  Datalsb = &H18
	  Datamsb = 0
       CASE 9600
	  Datalsb = &HC
	  Datamsb = 0
    END SELECT
    OUT DLL, Datalsb
    OUT DLM, Datamsb
    OUT LCR, &H3B
    OUT IER, 0
    OUT MCR, &H10
    OUT THR, &HAA
    Stime! = TIMER
    DO
    LOOP UNTIL (TIMER > Stime! + 1)
    IF INP(RBR) <> &HAA THEN
       Code = -NoCom
       EXIT SUB
    END IF
    OUT THR, &H55
    Stime! = TIMER
    DO
    LOOP UNTIL (TIMER > Stime! + 1)
    IF INP(RBR) <> &H55 THEN
       Code = -NoCom
       EXIT SUB
    END IF
    OUT MCR, 1
    Datalsb = INP(RBR)
END SUB

SUB Initials (command, Parameter&, Check, Code)
	     'Value  , Value     , Value, Reference
    ComResult = 0
    DevMsg = 0
    OUT LCR, &H2B
    Talk (address), ParameterBase
    OUT LCR, &H3B
    IF ComResult <> 0 THEN GOTO L1
'    IF Check = 1 THEN
'       Hibyte = (Parameter& / 256)
'       IF Hibyte <> ParameterBase THEN
'          ComResult = Badpara
'          GOTO L1
'       END IF
'    END IF
    lobyte = (Parameter& / 256 - Hibyte) * 256
    Talk (command), IdenNumber
    IF ComResult <> 0 THEN GOTO L1
    IF IdenNumber = Nack THEN
       ComResult = BadCom
       GOTO L1
    END IF
    Talk (Parameter&), header
    IF ComResult <> 0 THEN GOTO L1
    SELECT CASE header
	   CASE Errorfollows
		Talk (RTR), DevMsg
	   CASE MCUBusy
		ComResult = DevBusy
	   CASE Nack
		ComResult = Badpara
	   CASE ELSE
		IF (header AND &H80) = &H80 THEN Signed = 1 ELSE Signed = 0
		exponent = (header AND &H7C) / 4
		Temp = exponent AND &H10
		IF Temp = &H10 THEN
		   exponent = exponent AND &HF
		   exponent = exponent - 16
		END IF
		length = header AND &H3
    END SELECT
L1: Code = DevMsg - ComResult
END SUB

SUB ReadCom (Dataz) 'Reference
    Timeout% = 18 'About 1 second: This is the allotted delay for
		  'the Device to response
    Stime1! = TIMER
    Timerx = 0
    DO
    CHECKER = INP(LSR)
    Timerx = Timerx + 1
      DO
	 Stime2! = TIMER
      LOOP UNTIL ABS(Stime2! - Stime1!) > 0
      Stime1! = Stime2!
    LOOP UNTIL ((CHECKER AND &H11) = 1) OR (Timerx = Timeout%)
    IF Timerx <> Timeout% THEN Dataz = INP(RBR) ELSE ComResult = TimeErr
END SUB

'--------------------
' ReadData procedure
'--------------------
SUB ReadData (Parameter&, datax#, Code)
    DIM Temp!(0 TO 3), Temp1!(0 TO 3), Temp2#(0 TO 3)
    Temp1!(0) = 1
    Temp1!(1) = 256
    Temp1!(2) = 65536#
    Temp1!(3) = 16777216#
    Temp2#(0) = 256
    Temp2#(1) = 65536#
    Temp2#(2) = 16777216#
    Temp2#(3) = 4294967296#
    RG (ReadD), (Parameter&), 1, datax#, Code
    IF Code <> 0 THEN EXIT SUB
    Bufferl# = 0
    FOR i = 0 TO length
	Temp!(i) = Bufferb(i)
	Bufferl# = Bufferl# + Temp!(i) * Temp1!(i)
    NEXT i
    IF Signed THEN IF Bufferb(length) >= &H80 THEN Bufferl# = Bufferl# - Temp2#(length)
    'Bufferl& = CLNG(Bufferl#)
    datax# = Bufferl#
    datax# = datax# * EXP(CDBL(exponent) * Ln10#)
END SUB

'--------------------
' ReadFlag procedure
'--------------------
SUB ReadFlag (Parameter&, bit%, datax%, Code%)
	     'Value     , Val, Ref  , Ref
    IF (bit% < 0) OR (bit% > 31) OR (bit% > ((length% + 1) * 8 - 1)) THEN
       Code% = -BadData
       datax% = -1
       EXIT SUB
    END IF
    CALL ReadData((Parameter&), datad#, Code%)
    IF Code% <> 0 THEN
       datax% = -1
       EXIT SUB
    END IF
    mask% = 1
    IF (bit% > 0) AND (bit% < 8) THEN
       FOR i = 1 TO bit%
	   mask% = mask% * 2
       NEXT i
       datax% = Bufferb%(0) AND mask%
    END IF
    IF (bit% > 8) AND (bit% < 16) THEN
       FOR i = 9 TO bit%
	   mask% = mask% * 2
       NEXT i
       datax% = Bufferb%(1) AND mask%
    END IF
    IF (bit% > 16) AND (bit% < 24) THEN
       FOR i = 17 TO bit%
	   mask% = mask% * 2
       NEXT i
       datax% = Bufferb%(2) AND mask%
    END IF
    IF (bit% > 24) THEN
       FOR i = 25 TO bit%
	   mask% = mask% * 2
       NEXT i
       datax% = Bufferb%(3) AND mask%
    END IF
    IF bit% = 0 THEN datax% = Bufferb%(0) AND mask%
    IF bit% = 8 THEN datax% = Bufferb%(1) AND mask%
    IF bit% = 16 THEN datax% = Bufferb%(2) AND mask%
    IF bit% = 24 THEN datax% = Bufferb%(3) AND mask%
    IF (datax% <> 0) THEN datax% = 1
END SUB

SUB RG (command, Parameter&, Check, datax#, Code)
       'Value  , Value     , Value, Ref   , Ref
    DIM Temp&(0 TO 3)
    Initials (command), (Parameter&), (Check), Code
    IF Code <> 0 THEN
       datax# = 0
       EXIT SUB
    END IF
    FOR i = length TO 0 STEP -1
	Talk (RTR), Bufferb(i)
	Temp&(i) = Bufferb(i)
    NEXT i
    Bufferw&(0) = Temp&(0) + Temp&(1) * 256
    Bufferw&(1) = Temp&(2) + Temp&(3) * 256
END SUB

'----------------------
' SETUSERCOM procedure
'----------------------
SUB SetUserCom (addr%, Code%)
    IF (addr% < &H200) OR (addr% > &H3F8) THEN
       Code = -UserErr
       EXIT SUB
    END IF
    getaddress 0, addr%
END SUB

'---------------------------
' SWITCHTOADDRESS procedure
'---------------------------
SUB SwitchToAddress (A, Code%)
    IF (A < 0) OR (A > 255) THEN
       Code% = -AddrErr
       EXIT SUB
    END IF
    ComResult = 0
    address = A
END SUB

'-----------------------
' SWITCHTOCOM procedure
'-----------------------
SUB SwitchtoCom (Comx, Code%)
    ComResult% = 0
    Code% = 0
    IF (Comx < 1) OR (Comx > 5) THEN
       ComResult = ComnoErr
       Code% = -ComnoErr
       EXIT SUB
    END IF
    BaseAddress(1) = &H3F8
    BaseAddress(2) = &H2F8
    BaseAddress(3) = &H3E8
    BaseAddress(4) = &H2E8
    BaseAddress(5) = &H2E8
    IF Comx = 5 THEN
       getaddress 1, addrx%
       BaseAddress(5) = addrx%
    END IF
    THR = BaseAddress(Comx)
    RBR = BaseAddress(Comx)
    LSR = BaseAddress(Comx) + 5
    LCR = BaseAddress(Comx) + 3
    MCR = BaseAddress(Comx) + 4
    IER = BaseAddress(Comx) + 1
    DLL = BaseAddress(Comx)
    DLM = BaseAddress(Comx) + 1
    TRC = BaseAddress(Comx) + 7
END SUB

SUB Talk (Speak, Hear) 'Value , Ref
    WriteCom (Speak)
    ReadCom Hear
    IF ComResult <> 0 THEN
       ComResult = 0
       OUT LCR, &H2B
       WriteCom (address)
       ReadCom Hear
       OUT LCR, &H3B
       IF ComResult <> 0 THEN ComResult = TimeErr ELSE ComResult = DevBusy
    END IF
END SUB

SUB WG (Code)  'reference
    STATIC i, Answer
    FOR i = 0 TO length
	Talk (Bufferb(i)), Answer
	IF ComResult <> 0 THEN
	   Code = -ComResult
	   EXIT SUB
	END IF
	IF Answer = Nack THEN
	   ComResult = Badpara
	   Code = -ComResult
	   EXIT SUB
	END IF
    NEXT i
END SUB

SUB WriteCom (datax)
    OUT TRC, Transmit
    DO
       CHECKER = INP(LSR)
    LOOP UNTIL (CHECKER AND &H20) = &H20
    OUT THR, datax
    DO
       CHECKER = INP(LSR)
    LOOP UNTIL (CHECKER AND &H40) = &H40
    OUT TRC, Receive
END SUB

'---------------------
' WriteData procedure
'---------------------
SUB WriteData (Parameter&, datax#, Code)
	      'Value     , Value , Ref
    STATIC i, SignedRange#, UnsignedRange#, datad#
    DIM Hibyte#(0 TO 3), Temp2#(0 TO 3)
    Temp2#(0) = 128
    Temp2#(1) = 32768
    Temp2#(2) = 8388608#
    Temp2#(3) = 2147483648#
    Initials (WriteD), (Parameter&), 1, Code
    IF Code <> 0 THEN EXIT SUB
    datad# = datax#
    datad# = datad# * EXP(-exponent * Ln10)
    UnsignedRange# = Temp2#(length%) * 2#
    SignedRange# = Temp2#(length%)
    IF Signed THEN
       IF (datad# < SignedRange#) AND (datad# >= -SignedRange#) THEN
	     Bufferl& = CLNG(datad#)
	     DEF SEG = VARSEG(Bufferl&)
	     adx% = VARPTR(Bufferl&)
	     FOR i = 0 TO 3
	       Bufferb%(i) = PEEK(adx% + i)
	     NEXT i
       ELSE
	  ComResult = BadData
	  Code = -ComResult
	  EXIT SUB
       END IF
   ELSE
       IF (datad# >= 0) AND (datax# < UnsignedRange#) THEN
	  IF datad# >= 2147483648# THEN datad# = datad# - 4294967296#
	  Bufferl& = CLNG(datad#)
	  DEF SEG = VARSEG(Bufferl&)
	  adx% = VARPTR(Bufferl&)
	  FOR i = 0 TO 3
	      Bufferb%(i) = PEEK(adx% + i)
	  NEXT i
       ELSE
	  ComResult = BadData
	  Code = -ComResult
	  EXIT SUB
       END IF
  END IF
  WG Code
END SUB

