{$E+} {$N+}
UNIT DEVICES;
{JENCO Instruments Serial Communications Protocol}
{File  : Devices.pas     }
{Date  : 1992,1997}
{Pascal Implementation : RAGG }


INTERFACE
Uses CRT;
Type
   Array256 = array[0..255] of byte;
Procedure InitCom (ComNo: byte; BaudRate: word; var Code: integer);
Procedure SetUserCom(PortAddr:word; var Code: integer);
Procedure SwitchToCom(ComNo:byte; var Code: integer);
Procedure SwitchToAddress(a:byte);
Procedure ReadData(Parameter:word;var Data:double;var Code:integer);
Procedure ReadFlags(Parameter:word;var Data:longint;var Code:integer);
Procedure ReadFlag(Parameter:word;Bit:byte;var Data:integer;var Code:integer);
Procedure ReadMem(Addr:word;Len:byte;var Data:array256;var Code:Integer);
Procedure WriteData(Parameter:word;Data:Double; var Code:integer);
Procedure WriteMem(Addr:word;Len:byte;Data:array256;var Code:Integer);
procedure GetDevType (var DevType :Word; var Code: integer);

IMPLEMENTATION
var
devmsg :byte;

Type
	ComDataType	= 	record
   						case byte of
                     	1 : (L : longint);
                        2 : (B : array[0..3] of byte);
                        3 : (W : array[0..1] of word);
   					end;

Const
   RTR          = $D0;
   NAck         = $3E;
   Ack          = $00;
   ErrorFollows = $3C;
   MCUBusy      = $3D;
   Transmit     = 2  ;
   Receive      = 1  ;
   ReadD        = 1  ;		{ !!! originally 0 }
   WriteD       = 2  ;		{ !!! originally 1 }
   ReadE        = 0  ;	{ !!! originally 2 }
   WriteE       = 0  ;	{ !!! originally 3 }
   ReadM        = 128  ;	{ !!! originally 2 }
   WriteM       = 129  ;	{ !!! originally 3 }
{   Invoke_cmd   = 4  ; }	{ !! this command has been eliminated }
   Address       : byte = 0;
   BaseAddress   : array[1..5] of word = ($3F8,$2F8,$3E8,$2E8,$2E8);

const
   NO_ERROR       =  0;	{ everything is A OK }
   INVALID_COM    = -1; { invalid com nos. }
   INVALID_BAUD   = -2; { invalid baud rate }
   TIMED_OUT      = -3;	{ device did not respond within a certain time }
   DEVICE_BUSY    = -4; { device is busy, transaction not completed }
   UNKNOWN_PARAM  = -5; { parameter unknown to the device }
   UNKNOWN_CMD    = -6; { command unknown to the device }
   INVALID_DATA   = -7;	{ data is invalid, out of range }
   RESET_ERROR    = -8;	{ device did not reset }
   BAD_USER_COM   = -9;	{ user com port address is invalid }
   NO_COM         = -10;{ com# does not exist or is malfunctioning }
   ComResult      : integer = 0;
   TimeOut        : word    = 3000;
Var
   THR,RBR,LSR,LCR,
   MCR,IER,DLL,DLM,TRC : word;
   Header,Length       : byte;
   Exponent            : shortint;
   Signed              : boolean;
   Buffer              : ComDataType;
   ParameterBase,
   IdenNumber          : byte;


   
(*****************************************************************************
	Function:
		Sets the base address for Com5
	Input Parameters:
		PortAddr = $200 to $3F8
	Returns:
		NO_ERROR if Com5 set properly
		INVALID_ADDR if the Addr is not within 0x200 to 0x3F8
	Warning:
		The Addr is limited to this range, since anything below 0x200 is
		used by the main system board, while 0x3F8 is the highest address
		available for most serial cards, any port address within this range
		that conflicts with other cards may cause damage to the computer
		or its peripherals. The user is cautioned to use his own discretion
		in selecting an unused port address for the serial card.
*****************************************************************************)
Procedure SetUserCom(PortAddr:word; var Code: integer);
begin
	if (PortAddr < $200) or (PortAddr > $3F8) then
		begin
			ComResult := BAD_USER_COM;
			Code := ComResult;
		end
	else
		begin
	  		BaseAddress[5] := PortAddr;
			ComResult := NO_ERROR;
			Code := ComResult;
		end;
end;

(*****************************************************************************
	Input Parameters:
		ComNo = 1 to 5
   Returns:
		NO_ERROR if ComNo is within 1 to 5
		INVALID_COM if the ComNo is not supported.
*****************************************************************************)
Procedure SwitchToCom (ComNo: byte; var Code: integer);
begin
	if (ComNo < 1) or (ComNo > 5) then
   	begin
     		ComResult := INVALID_COM;
     		Code := ComResult;
	   end
	else
		begin
			ComResult := NO_ERROR;
			Code := ComResult;
			THR := BaseAddress[ComNo] + 0;
			RBR := BaseAddress[ComNo] + 0;
			LSR := BaseAddress[ComNo] + 5;
			LCR := BaseAddress[ComNo] + 3;
			MCR := BaseAddress[ComNo] + 4;
			IER := BaseAddress[ComNo] + 1;
			DLL := BaseAddress[ComNo] + 0;
			DLM := BaseAddress[ComNo] + 1;
			TRC := BaseAddress[ComNo] + 7;
		end;
end;

(*****************************************************************************
	Function:
		Initializes the ComNo to the given BaudRate and the other parameters
		needed by the Jenco Serial Communications Protocol
	Input Parameters:
		ComNo 	= 1 to 5
		BaudRate = 300, 600, 1200, 2400, 4800, 9600
   Returns:
		NO_ERROR if initialization was successful,
		INVALID_COM if the ComNo is not supported,
		INVALID_BAUD if the BaudRate is not supported.
	Warning:
		This must be called first before any of the other routines is called
		in this module, otherwise unpredictable behavior will occur.
*****************************************************************************)
Procedure InitCom (ComNo: byte; BaudRate:word; var Code: integer);
	var
   	DataLSB,DataMSB : byte;
   begin
		ComResult := NO_ERROR;
		Code := ComResult;
		SwitchToCom(ComNo, Code);
      if ComResult <> NO_ERROR then
			begin
				Code := ComResult;
				exit;
			end;
     	port[LCR] := $BB;
      case BaudRate of
         300 : begin
                  DataLSB := $80;
                  DataMSB := $01;
               end;
         600 : begin
                  DataLSB := $C0;
                  DataMSB := $00;
               end;
        1200 : begin
                  DataLSB := $60;
                  DataMSB := $00;
               end;
        2400 : begin
                  DataLSB := $30;
                  DataMSB := $00;
               end;
        4800 : begin
                  DataLSB := $18;
                  DataMSB := $00;
               end;
        9600 : begin
                  DataLSB := $0C;
                  DataMSB := $00;
               end;
        else   begin
	          ComResult := INVALID_BAUD;
                  Code := ComResult;
               end;
      end;
        port[DLL] := DataLSB;
        port[DLM] := DataMSB;
        port[LCR] := $3B;
        port[IER] := $00;
		(* perform the loopback test to see if the Com really exists *)
		port [MCR] := $10;	{ set the UART to perform loopback }
		port [THR] := $AA;
		delay (40);
		if port [RBR] <> $AA then begin
			ComResult := NO_COM;
			Code := ComResult;
			exit;
		end;
		port [THR] := $55;
		delay (40);
		if port [RBR] <> $55 then begin
			ComResult := NO_COM;
			Code := ComResult;
			exit;
		end;
      port[MCR] := $01;
   end;

Procedure ReadCom(var Data : byte);
   var
   Checker : byte;
   Timer   : word;
   begin
   	Timer := 0;
        repeat
            Checker := port[LSR];
            Inc(Timer);
            delay(1);
        until (Checker and $01 = $01) or (Timer = TimeOut);
        if Timer <> Timeout
      	then Data  := port [RBR]
        else ComResult := TIMED_OUT;
   end;

Procedure WriteCom(Data : byte);
	var
   	Checker : byte;
   begin
   	port[TRC] := Transmit;
      repeat
      	Checker := port[LSR];
      until (Checker and $20 = $20);
      port[THR] := Data;
      repeat
      	Checker := port[LSR];
      until Checker and $40 = $40;
      port[TRC] := Receive;
   end;

Procedure Talk(Speak:byte; var Hear:byte);
   begin
      WriteCom(Speak);
      ReadCom (Hear);
      if ComResult <> 0 then
      	begin
            ComResult := NO_ERROR;
            port[LCR] := $2B;
            WriteCom  (Address);
            ReadCom   (Hear);
            port[LCR] := $3B;
            if ComResult <> 0
               then ComResult := TIMED_OUT
               else ComResult := DEVICE_BUSY;
        end;
   end;

Procedure Initials(Command:byte;Parameter:word;Check:boolean;
						 var Code:integer);
   label
   	l1;
	var
		DevMsg: byte;
		Param : byte;
   begin
      ComResult := NO_ERROR;
      DevMsg    := 0;
      port[LCR] := $2B;
      Talk(Address,ParameterBase);
      port[LCR] := $3B;
      if ComResult <> 0 then goto l1;
		{ !!! the check for parameter base is no longer valid,
			it has been replaced by JCP version no. }
		{
      if Check = true then
         begin
            if hi(Parameter) <> ParameterBase then
               begin
                   ComResult := UNKNOWN_PARAM;
                   goto l1;
               end;
         end;
		}
      Param := lo(Parameter);
      Talk(Command,IdenNumber);
      if ComResult <> NO_ERROR then goto l1;
      if IdenNumber = NAck then
         begin
             ComResult := UNKNOWN_CMD;
             goto l1;
         end;
      Talk(Param,Header);
      if ComResult <> NO_ERROR then goto l1;
      case Header of
           ErrorFollows : Talk(RTR,DevMsg);
           MCUBusy      : ComResult := DEVICE_BUSY;
           NAck         : ComResult := UNKNOWN_PARAM;
      else
           begin
               if (Header and $80) = $80
                  then Signed := true
                  else Signed := false;
               Exponent := (Header and $7C) shl 1;
               Exponent := Exponent div 8;
               Length   := Header and $03;
          end;
      end;
l1:
     Code := ComResult+DevMsg;
   end;

Procedure RG(Command:byte;Parameter:word;Check:boolean;
				 var Data:double;var Code:integer);
   var  i : shortint;
   begin
   	Initials(Command,Parameter,Check,Code);
      if Code <> NO_ERROR then
      	begin
         	Data := 0;
            exit;
         end;
      Buffer.L := 0;
		for i := Length downto 0 do begin
			Talk(RTR,Buffer.B[i]);
			if ComResult <> NO_ERROR then begin
				Code := ComResult;
				exit;
			end;
		end;
   end;

Procedure ReadData(Parameter:word;var Data:double;var Code:integer);
   label
   	l1;
	var  i : byte;
   begin
   	RG(ReadD,Parameter,true,Data,Code);
      if Code <> NO_ERROR then exit;
      if Signed
      	then
         	begin
            	if Buffer.B[Length] >= $80 then
               	for  i := Length+1 to 3 do Buffer.B[i] := $FF;
               Data := Buffer.L;
            end
         else Data := Buffer.W[0] + Buffer.W[1]*65536.0;
      Data := Data * exp(Exponent*ln(10));
   end;

Procedure ReadMem    ( Addr   : word;
                       Len    : byte;
                       var
                       Data   : array256;
                       var
                       Code   :  Integer);
	var
   i,param   : byte;
   parameter : word;
   Label L1;
	begin
      Parameter := 128;
      ComResult := NO_ERROR;
      DevMsg    := 0;
      port[LCR] := $2B;
      Talk(Address,ParameterBase);
      port[LCR] := $3B;
      if ComResult <> 0 then goto l1;
      Param := lo(Parameter);
      Talk (ReadM,IdenNumber);
      if ComResult <> NO_ERROR then goto l1;
      if IdenNumber = NAck then
         begin
             ComResult := UNKNOWN_CMD;
             exit;
         end;
		Talk (lo(Addr),IdenNumber);
      if ComResult <> NO_ERROR then goto l1;
		Talk (hi(Addr), Header);
      if ComResult <> NO_ERROR then goto l1;
      case Header of
           ErrorFollows : Talk(RTR,DevMsg);
           MCUBusy      : ComResult := DEVICE_BUSY;
           NAck         : ComResult := UNKNOWN_PARAM;
      end;
		Talk (Len,IdenNumber);
      if ComResult <> NO_ERROR then goto l1;
		for i := 0 to Len do
        begin
			  Talk (RTR, Data [i]);
			  if ComResult <> NO_ERROR then
              begin
				     Code := ComResult;
                 exit;
              end;
        end;
l1:  Code := ComResult+DevMsg;
   end;

Procedure ReadFlags(Parameter:word;var Data:longint;var Code:integer);
	var
		DataD : double;
   begin
   	ReadData(Parameter,DataD,Code);
      if Code <> NO_ERROR
			then Data := 0
         else Data := Buffer.L;
   end;

Procedure ReadFlag(Parameter:word;Bit:byte;var Data:integer;var Code:integer);
	var
		mask  : word;
      i     : byte;
      DataD : double;
   begin
   	ReadData(Parameter,DataD,Code);
      if Code <> NO_ERROR then
      	begin
         	Data := -1;
            exit;
         end;
      mask := 1;
      if bit =0 then DataD:=Buffer.w[0] and 1;
      if (bit > 0) and (bit < 16) then
          begin
             for i := 1 to bit do mask := mask * 2;
             DataD := Buffer.w[0] and mask;
          end;
      if bit>16 then
         begin
            for i := 16 to bit do mask := mask * 2;
            DataD := Buffer.w[1] and mask;
         end;
      Data := round(DataD/mask);
   end;

Procedure WG(var Code:integer);
	var
		i      : shortint;
      Answer : byte;
   begin
   	for i := 0 to Length do
      	begin
         	Talk(Buffer.B[i],Answer);
            if ComResult <> NO_ERROR then
            	begin
               	Code := ComResult;
                  exit;
               end;
            if Answer = NAck then
            	begin
						ComResult := INVALID_DATA;
                  Code := ComResult;
               	exit;
               end;
         end;
   end;

Procedure WriteData(Parameter:word;Data:double;var Code:integer);
	var
   	i                               : shortint;
      SignedRange,UnsignedRange,DataD : double;
   begin
   	Initials(WriteD,Parameter,true,Code);
      if Code <> NO_ERROR then exit;
      DataD := Data;
      DataD := DataD * exp(-Exponent*ln(10));
      UnsignedRange := exp((Length+1)*ln(256));
      SignedRange:=UnsignedRange/2;
      If Signed then
         begin
            if (DataD < SignedRange) and (DataD>=-SignedRange)
               	then Buffer.L := round(DataD)
                else
                   begin
                      ComResult := INVALID_DATA;
                      Code := ComResult;
                      exit;
                   end;
         end
      else
         begin
            if (DataD >= 0) and (Data < UnsignedRange) then
               begin
                  if DataD >= 2147483648.0
                     then DataD := DataD - 4294967296.0;
                  Buffer.L := round(DataD);
               end
            else
               begin
                  ComResult := INVALID_DATA;
                  Code := ComResult;
                  exit;
               end;
         end;
      WG(Code);
   end;

Procedure WriteMem ( Addr           : word;
                     Len            : byte;
                     Data           : array256;
                     var
                     Code           : integer);
	var
		i         : integer;
		DevMsg,
		Param,
		Reply     : byte;
      Parameter : word;
      label L1;
   Begin
      Parameter := 129;
      ComResult := NO_ERROR;
      DevMsg    := 0;
      port[LCR] := $2B;
      Talk(Address,ParameterBase);
      port[LCR] := $3B;
      if ComResult <> 0 then goto L1;
      Param := lo(Parameter);
      Talk(Parameter,IdenNumber);
      if ComResult <> NO_ERROR then goto L1;
      if IdenNumber = NAck then
         begin
             ComResult := UNKNOWN_CMD;
             Code :=ComResult;
             exit;
         end;
      Talk (lo(Addr),Header);
      if ComResult <> NO_ERROR then goto L1;
      Talk (hi(Addr),Header);
      if ComResult <> NO_ERROR then goto L1;
      case Header of
           ErrorFollows : Talk(RTR,DevMsg);
           MCUBusy      : ComResult := DEVICE_BUSY;
           NAck         : ComResult := UNKNOWN_PARAM;
      end;
      Talk (Len,Header);
      if ComResult <> NO_ERROR then goto L1;
		for i := 0 to Len do
          begin
			    Talk (Data [i], Reply);
			    if ComResult <> NO_ERROR then
                begin
				       Code := ComResult;
				       exit;
			       end;
          end;
   L1:    Code:=ComResult + DevMsg;
   End;


(*****************************************************************************
	Input Parameters:
		none
	Returns:
		The device type at the current device address
*****************************************************************************)
procedure GetDevType (var DevType :Word; var Code: integer);
var
	Hitype,trycmd, reply: byte;

begin
	trycmd := 0;
   Code   :=0;
	repeat
             port[LCR] := $2B;
             Talk(Address, hitype);
             if ComResult <> NO_ERROR then
                begin
                     Code := ComResult;
                     exit;
                end;
             DevType := Hitype shl 8;
             port[LCR] := $3B;
             Talk (trycmd, reply);
             if reply <> NACK then
                begin
                    DevType := DevType or reply;
                    exit;
                end;
             trycmd := trycmd + 1;
        until trycmd > 5;
        Code := UNKNOWN_CMD;
eND;

Procedure SwitchToAddress(a:byte);
	begin
   	ComResult := NO_ERROR;
      Address   := a;
   end;

{**************************************************************************
	Input Parameters:
		Cmd - the command, specific to the given device
	Returns: Code
		= NO_ERROR if the command was executed successfully
		= UNKNOWN_CMD if the command is unrecognized and ignored by the device
		= TIMED_OUT if the device did not respond within a certain period
		= DEVICE_BUSY if the device cannot process the command immediately
***************************************************************************}
{procedure Invoke (cmd:integer; var Code :integer);
   var
	     reply : byte;
   const
        SPECIAL_CMDS = 8;
        OFFSET       = 16;
   Begin
        port[LCR] := $2B;
        Talk(Address, reply);
        port[LCR] := $3B;
	IF (Cmd < SPECIAL_CMDS) then } 	{0 to 7 are special commands }
          { begin
              Talk(Cmd+OFFSET,reply);}  { offset for the first 16 reserved }
              {if (ComResult <> NO_ERROR) then
                 begin
                    Code:= ComResult;
                    exit;
                 end;
              if (reply = NACK) then
                 begin
                    Code := UNKNOWN_CMD;
                    exit;
                 end;
           end
	ELSE
           begin
              Talk (INVOKE_CMD,reply);
              if (ComResult <> NO_ERROR) then
                 Begin
                    Code:=ComResult;
                    exit;
                 end;
              if (reply = NACK) then
                 Begin
                    Code:=UNKNOWN_CMD;
                    Exit;
                 end;
              Talk(Cmd-SPECIAL_CMDS,reply);
              if (Comresult<> NO_ERROR) then
                 begin
                    Code:=UNKNOWN_CMD;
                    exit;
                 end;
              if (reply = NACK) then
                 begin
                    Code:= UNKNOWN_PARAM;
                    exit;
                 end;
        end;
        Code:= NO_ERROR;
   end;}

end.
